/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: val.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"



/*****************\
*                 *
*  Type checking  *
*                 *
\*****************/

void type_check(NODEPTR e, unsigned type)
{
    assert2 (node_type(e) == type,
            "found expression of type %s when expecting %s",
             print_name (e), Keyword[type-MINTOKEN]);
}

/******************************************\
*                                         *
*  checking  whether e is static variable *
*                                         *
\******************************************/

void static_check(NODEPTR e)
{
    assert1 (node_type(e) == STATIC,
            "variable %s is of type state should be static",
             node_nm_name(e));
}

/********************************\
*                                *
*  Test whether val is defined   *
*                                *
\********************************/
int val_defined(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:
        case LAMBDA:
            return (T);
        case iLAMBDA:
            return (ilambda_defined (e)); 
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (list_defined (e));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}

int prev_val_defined(NODEPTR e, unsigned int k)
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA: 
            return (T);
        case iLAMBDA:
            return (prev_ilambda_defined (e,k));
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (prev_list_defined (e,k));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}

int curr_val_defined(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA:
            return (T);
        case iLAMBDA:
          return (curr_ilambda_defined (e));
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (curr_list_defined (e));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}

int next_val_defined(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA:
            return (T);
        case iLAMBDA:
            return (next_ilambda_defined (e));
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (next_list_defined (e));
        default:
            error_break1 ("value cannot be of type %s", print_name(e));
            return (F);
    }
}

int val_defined2(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA:
            return (T);
        case iLAMBDA:
            return (ilambda_defined (e));
        case LOC:
            return (pointer_defined ((e)));
        case LIST:
            return (list_defined (e));
        default:
            return (F);
    }
}

int pointer_defined(NODEPTR p)
{

  /*printf("node_loc_tonext=%d\n",node_loc_tonext(p));*/
  /*printf("pointer_defined=%d\n",(mem_defined1 (node_loc_tonext(p) ?
                         next_loc (node_loc_ptr(p)) :
                         curr_loc (node_loc_ptr(p)))));*/
  
    return (mem_defined1 (node_loc_tonext(p) ?
                         next_loc (node_loc_ptr(p)) :
                         curr_loc (node_loc_ptr(p))));
}


/* check whether an ilambda is defined */

int ilambda_defined(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR env, id;
  NODEPTR bnd_env, bnd_blk_base_loc;
  
  
  if (bnd_1 == NULL ) {
    return (T);
  } else {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
        /*printf("ilambda_defined: !mem_defined\n");*/
        return (F);
      }
      if (! val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
        /*printf("ilambda_defined: !val_defined\n");*/
        return (F);
      }
    }
    bnd_1 = node_bnd_next(bnd_1);
    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
          /*printf("ilambda_defined: !mem_defined\n");*/
          return (F);
        }
        if (! val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
          /*printf("ilambda_defined: !val_defined\n");*/
          return (F);
        }
      }
      bnd_1 = node_bnd_next(bnd_1);
    }
    return (T);  
  }
  
}

int prev_ilambda_defined(NODEPTR ilambda, unsigned int k)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR env, id;
  NODEPTR bnd_env, bnd_blk_base_loc;
  
  
  if (bnd_1 == NULL ) {
    return (T);
  } else {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
        /*printf("prev_ilambda_defined: !mem_defined\n");*/
        return (F);
      }
      if (! prev_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)), k) ) {
        /*printf("prev_ilambda_defined: !prev_val_defined\n");*/
        return (F);
      }
    }
    bnd_1 = node_bnd_next(bnd_1);
    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
          /*printf("prev_ilambda_defined: !mem_defined\n");*/
          return (F);
        }
        if (! prev_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)),k)) {
          /*printf("prev_ilambda_defined: !prev_val_defined\n");*/
          return (F);
        }
      }
      bnd_1 = node_bnd_next(bnd_1);
    }
    return (T);  
  }
  
}

int curr_ilambda_defined(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR env, id;
  NODEPTR bnd_env, bnd_blk_base_loc;
  
  
  if (bnd_1 == NULL ) {
    return (T);
  } else {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
        /*printf("curr_ilambda_defined: !mem_defined\n");*/
        return (F);
      }
      if (! curr_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
        /*printf("curr_ilambda_defined: !curr_val_defined\n");*/
        return (F);
      }
    }
    bnd_1 = node_bnd_next(bnd_1);
    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
          /*printf("curr_ilambda_defined: !mem_defined\n");*/
          return (F);
        }
        if (! curr_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
          /*printf("curr_ilambda_defined: !curr_val_defined\n");*/
          return (F);
        }
      }
      bnd_1 = node_bnd_next(bnd_1);
    }
    return (T);  
  }
  
}

int next_ilambda_defined(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR env, id;
  NODEPTR bnd_env, bnd_blk_base_loc;
  
  
  if (bnd_1 == NULL ) {
    return (T);
  } else {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
        /*printf("next_ilambda_defined: !mem_defined\n");*/
        return (F);
      }
      if (! next_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
        /*printf("next_ilambda_defined: !next_val_defined\n");*/
        return (F);
      }
    }
    bnd_1 = node_bnd_next(bnd_1);
    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        if (! mem_defined1 (bnd_blk_base_loc + node_nm_off(id) ) ) {
          /*printf("next_ilambda_defined: !mem_defined\n");*/
          return (F);
        }
        if (! next_val_defined (get_mem(bnd_blk_base_loc + node_nm_off(id)))) {
          /*printf("next_ilambda_defined: !next_val_defined\n");*/
          return (F);
        }
      }
      bnd_1 = node_bnd_next(bnd_1);
    }
    return (T);  
  }
  
}



/* Check whether list is defined. */
int list_defined(NODEPTR list)
{
    register int     current_size;
    register unsigned i;
    register NODEPTR base_loc;
    NODEPTR size_loc;
    
    if (list_size_defined (list)) {
        base_loc = list_base (list);
        size_loc = list_size (list);
        current_size = get_int (size_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined1 (adj_loc1(base_loc + offset2 (i,node_mem_history(size_loc))))) {
                return (F);
            }
            if (!val_defined (get_mem (adj_loc1(base_loc + offset2 (i,node_mem_history(size_loc)))))) {
                return (F);
            }
        }
        return (T);
    }
    else{
        return (F);
    }
}

int prev_list_defined(NODEPTR list, unsigned int k)
{
    register int     current_size;
    register unsigned i;
    register NODEPTR base_loc,pbase_loc;
    NODEPTR size_loc,psize_loc;
    
    if (prev_list_size_defined (list,k)) {
        base_loc = list_base (list);
        size_loc = list_size (list);
        pbase_loc = prev_loc1(base_loc,k);
        psize_loc = prev_loc1(size_loc,k);
        current_size = get_int (psize_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined1 (pbase_loc + offset2 (i,node_mem_history(size_loc)))) {
                return (F);
            }
            if (!prev_val_defined (get_mem (pbase_loc + offset2 (i,node_mem_history(size_loc))),k)) {
                return (F);
            }
        }
        return (T);
    }
    else{
        return (F);
    }
}

int curr_list_defined(NODEPTR list)
{
    register int     current_size;
    register unsigned i;
    register NODEPTR base_loc,cbase_loc;
    NODEPTR size_loc,csize_loc;
    
    if (curr_list_size_defined (list)) {
        base_loc = list_base (list);
        size_loc = list_size (list);
        cbase_loc = curr_loc(base_loc);
        csize_loc = curr_loc(size_loc);
        current_size = get_int (csize_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined1 (cbase_loc + offset2 (i,node_mem_history(size_loc)))) {
                return (F);
            }
            if (!curr_val_defined (get_mem (cbase_loc + offset2 (i,node_mem_history(size_loc))))) {
                return (F);
            }
        }
        return (T);
    }
    else{
        return (F);
    }
}

int next_list_defined(NODEPTR list)
{
    register int     current_size;
    register unsigned i;
    register NODEPTR base_loc,nbase_loc;
    NODEPTR size_loc,nsize_loc;
    
    if (next_list_size_defined (list)) {
        base_loc = list_base (list);
        size_loc = list_size (list);
        nbase_loc = next_loc(base_loc);
        nsize_loc = next_loc(size_loc);
        current_size = get_int (nsize_loc);
        for (i = 0; (int)i < current_size; i += 1) {
            if (!mem_defined1 (nbase_loc + offset2 (i,node_mem_history(size_loc)))) {
                return (F);
            }
            if (!next_val_defined (get_mem (nbase_loc + offset2 (i,node_mem_history(size_loc))))) {
                return (F);
            }
        }
        return (T);
    }
    else{
        return (F);
    }
}

int max_size(NODEPTR list)
{
    NODEPTR size_loc;
    if (mem_defined1 (size_loc = list_size (list)) && is_static_mem (size_loc))
        return (get_int (size_loc));
    else
        return (node_blk_size(node_lst_blk(list)) - 1);
}

int prev_max_size(NODEPTR list, unsigned int k)
{
    NODEPTR size_loc,psize_loc;;
    size_loc = list_size (list);
    psize_loc = prev_loc1(size_loc,k);
    
    if (mem_defined1 (psize_loc) && is_static_mem (psize_loc))
        return (get_int (psize_loc));
    else
        return (node_blk_size(node_lst_blk(list)) - 1);
}


int curr_max_size(NODEPTR list)
{
    NODEPTR size_loc,csize_loc;;
    size_loc = list_size (list);
    csize_loc = curr_loc(size_loc);
    
    if (mem_defined1 (csize_loc) && is_static_mem (csize_loc))
        return (get_int (csize_loc));
    else
        return (node_blk_size(node_lst_blk(list)) - 1);
}

int next_max_size(NODEPTR list)
{
    NODEPTR size_loc,nsize_loc;;
    size_loc = list_size (list);
    nsize_loc = next_loc(size_loc);
    
    if (mem_defined1 (nsize_loc) && is_static_mem (nsize_loc))
        return (get_int (nsize_loc));
    else
        return (node_blk_size(node_lst_blk(list)) - 1);
}

/* Check whether size of list is defined. */
int list_size_defined(NODEPTR list)
{
    int     current_size;
    register NODEPTR size_loc = list_size (list);
    
    if (mem_defined1 (size_loc)) {
        current_size = get_int (size_loc);
        assert2d ((current_size >= 0) && (current_size <= max_size (list)),
                "current size of list is out of range [0,%d]: %d",
                max_size (list), current_size);
        return (T);
    }
    else {
        return (F);
    }
}

/* Check whether size of list is defined. */
int prev_list_size_defined(NODEPTR list, unsigned int k)
{
    int     current_size;
    register NODEPTR psize_loc,size_loc;
    size_loc = list_size (list);
    psize_loc = prev_loc1(size_loc,k);
    
    
    if (mem_defined1 (psize_loc)) {
        current_size = get_int (psize_loc);
        assert2d ((current_size >= 0) && (current_size <= prev_max_size (list,k)),
                "current size of list is out of range [0,%d]: %d",
                  prev_max_size (list,k), current_size);
        return (T);
    }
    else {
        return (F);
    }
}

/* Check whether size of list is defined. */
int curr_list_size_defined(NODEPTR list)
{
    int     current_size;
    register NODEPTR size_loc, csize_loc;
    size_loc = list_size (list);
    csize_loc = curr_loc(size_loc);
    
    if (mem_defined1 (csize_loc)) {
        current_size = get_int (csize_loc);
        assert2d ((current_size >= 0) && (current_size <= curr_max_size (list)),
                "current size of list is out of range [0,%d]: %d",
                curr_max_size (list), current_size);
        return (T);
    }
    else {
        return (F);
    }
}

/* Check whether size of list is defined. */
int next_list_size_defined(NODEPTR list)
{
    int     current_size;
    register NODEPTR size_loc, nsize_loc;
    size_loc = list_size (list);
    nsize_loc = next_loc(size_loc);
    
    if (mem_defined1 (nsize_loc)) {
        current_size = get_int (nsize_loc);
        assert2d ((current_size >= 0) && (current_size <= next_max_size (list)),
                "current size of list is out of range [0,%d]: %d",
                next_max_size (list), current_size);
        return (T);
    }
    else {
        return (F);
    }
}


/********************************************************\
*                                                        *
*  Copying, comparing, assigning and outputting values.  *
*                                                        *
\********************************************************/

/* convert a value to a printable form */
void output_val(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (node_ival(e), 0);
            break;
        case STRING:
            output_char ('"');
            output_str (node_strval(e));
            output_char ('"');
            break;
        case FLOAT:
            /* output_char ('$');*/
            output_fl (node_flval(e));
            /* output_char ('$');*/
            break;
        case MPFR:
            /* output_str ("$$");*/
            output_mpfr(node_mpfrval(e));
            /* output_str ("$$");*/
            break; 
        case MPZ:
            output_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
            output_str( "-0z");
            output_mpz_abs( node_mpzval(e));
          } else{
            output_str ( "0z" );
            output_mpz (node_mpzval(e));
            }*/
            break;
        case MPX:
            output_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;  
        case LIST:
            output_list (e);
            break;
        case LOC:
            output_char ('@');
            output_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;
        case iLAMBDA:
          /*
            output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");
          */
            output_ilambda(e);
            break;
        default:
            error_break1 ("output_val: expression of type %s cannot be output",
                    print_name (e));
    }
}

void output_val_prev(NODEPTR e, unsigned int k)
{
    switch (node_type(e)) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (node_ival(e), 0);
            break;
        case STRING:
            output_char ('"');
            output_str (node_strval(e));
            output_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            output_fl (node_flval(e));
            /* output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            output_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;
        case MPZ:
            output_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
             } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
       case MPX:
            output_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;
        case LIST:
            output_list_prev (e,k);
            break;
        case LOC:
            output_char ('@');
            output_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;
        case iLAMBDA:
          /*output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");
          */
            output_ilambda_prev(e,k); 
            break;    
        default:
            error_break1 ("output_val_prev: expression of type %s cannot be output",
                    print_name (e));
    }
}

void foutput_val_prev(NODEPTR e, unsigned int k)
{
    switch (node_type(e)) {
        case NIL:
	    foutput_str ("nil");
	    break;
        case BOOL:
            foutput_str (e == True ? "true" : "false");
            break;
        case INT:
            foutput_int (node_ival(e), 0);
            break;
        case STRING:
            foutput_char ('"');
            foutput_str (node_strval(e));
            foutput_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            foutput_fl (node_flval(e));
            /* output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            foutput_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;
        case MPZ:
            foutput_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
             } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
       case MPX:
            foutput_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;
        case LIST:
            foutput_list_prev (e,k);
            break;
        case LOC:
            foutput_char ('@');
            foutput_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            foutput_lambda (e);
            break;
        case iLAMBDA:
          /*output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");
          */
            foutput_ilambda_prev(e,k); 
            break;    
        default:
            error_break1 ("foutput_val_prev: expression of type %s cannot be output",
                    print_name (e));
    }
}


void output_val_next(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (node_ival(e), 0);
            break;
        case STRING:
            output_char ('"');
            output_str (node_strval(e));
            output_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            output_fl (node_flval(e));
            /*output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            output_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;     
        case MPZ:
            output_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
            } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
        case MPX:
            output_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;    
        case LIST:
            output_list_next (e);
            break;
        case LOC:
            output_char ('@');
            output_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;
        case iLAMBDA:
          /*
            output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");
          */
            output_ilambda_next(e);
            break;    
        default:
            error_break1 ("output_val_next: expression of type %s cannot be output",
                    print_name (e));
    }
}

void foutput_val_next(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    foutput_str ("nil");
	    break;
        case BOOL:
            foutput_str (e == True ? "true" : "false");
            break;
        case INT:
            foutput_int (node_ival(e), 0);
            break;
        case STRING:
            foutput_char ('"');
            foutput_str (node_strval(e));
            foutput_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            foutput_fl (node_flval(e));
            /*output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            foutput_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;     
        case MPZ:
            foutput_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
            } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
        case MPX:
            foutput_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;    
        case LIST:
            foutput_list_next (e);
            break;
        case LOC:
            foutput_char ('@');
            foutput_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            foutput_lambda (e);
            break;
        case iLAMBDA:
          /*
            output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");
          */
            foutput_ilambda_next(e);
            break;    
        default:
            error_break1 ("foutput_val_next: expression of type %s cannot be output",
                    print_name (e));
    }
}



void output_val_curr(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    output_str ("nil");
	    break;
        case BOOL:
            output_str (e == True ? "true" : "false");
            break;
        case INT:
            output_int (node_ival(e), 0);
            break;
        case STRING:
            output_char ('"');
            output_str (node_strval(e));
            output_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            output_fl (node_flval(e));
            /*output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            output_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;     
        case MPZ:
            output_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
            } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
        case MPX:
            output_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;    
        case LIST:
            output_list_curr (e);
            break;
        case LOC:
            output_char ('@');
            output_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            output_lambda (e);
            break;
        case iLAMBDA:
          /*output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");*/
            output_ilambda_curr(e);
            break;    
        default:
            error_break1 ("output_val_curr: expression of type %s cannot be output",
                    print_name (e));
    }
}

void foutput_val_curr(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    foutput_str ("nil");
	    break;
        case BOOL:
            foutput_str (e == True ? "true" : "false");
            break;
        case INT:
            foutput_int (node_ival(e), 0);
            break;
        case STRING:
            foutput_char ('"');
            foutput_str (node_strval(e));
            foutput_char ('"');
            break;
        case FLOAT:
            /*output_char ('$');*/
            foutput_fl (node_flval(e));
            /*output_char ('$');*/
            break;
        case MPFR:
            /*output_str ("$$");*/
            foutput_mpfr(node_mpfrval(e));
            /*output_str ("$$");*/
            break;     
        case MPZ:
            foutput_mpz(node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              output_str( "-0z");
              output_mpz_abs( node_mpzval(e));
            } else{
              output_str ( "0z" );
              output_mpz (node_mpzval(e));
              }*/
            break;
        case MPX:
            foutput_mpx(node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            output_str( "-0x");
            output_mpx_abs( node_mpxval(e));
          } else{
            output_str ( "0x" );
            output_mpx (node_mpxval(e));
            }*/
            break;    
        case LIST:
            foutput_list_curr (e);
            break;
        case LOC:
            foutput_char ('@');
            foutput_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            foutput_lambda (e);
            break;
        case iLAMBDA:
          /*output_str("( ");
            output_lambda (node_lnk_1(e));
            output_str(" with ");
            output_expn(node_lnk_2(e),-1);
            output_str(" )");*/
            foutput_ilambda_curr(e);
            break;    
        default:
            error_break1 ("foutput_val_curr: expression of type %s cannot be output",
                    print_name (e));
    }
}


void foutput_val(NODEPTR e)
{
    switch (node_type(e)) {
        case NIL:
	    foutput_str ("nil");
	    break;
        case BOOL:
            foutput_str (e == True ? "true" : "false");
            break;
        case INT:
            foutput_int (node_ival(e), 0);
            break;
        case STRING:
            foutput_char ('"');
            foutput_str (node_strval(e));
            foutput_char ('"');
            break;
        case FLOAT:
            /*foutput_char ('$');*/
            foutput_fl (node_flval(e));
            /*foutput_char ('$');*/
            break;
        case MPFR:
            /*foutput_str ("$$");*/
            foutput_mpfr(node_mpfrval(e));
            /*foutput_str ("$$");*/
            break;     
        case MPZ:
            foutput_mpz (node_mpzval(e));
            /*if (mpz_sgn(node_mpzval(e)) == -1) {
              foutput_str( "-0z");
              foutput_mpz_abs( node_mpzval(e));
            } else{
              foutput_str ( "0z" );
              foutput_mpz (node_mpzval(e));
              }*/
            break;
         case MPX:
            foutput_mpx (node_mpxval(e));
            /*if (mpz_sgn(node_mpxval(e)) == -1) {
            foutput_str( "-0x");
            foutput_mpx_abs( node_mpxval(e));
          } else{
            foutput_str ( "0x" );
            foutput_mpx (node_mpxval(e));
            }*/
            break;   
        case LIST:
            foutput_list (e);
            break;
        case LOC:
            foutput_char ('@');
            foutput_int ((int64_t)cell_no (node_loc_ptr(e)), 0);
            break;
        case LAMBDA:
            foutput_lambda (e);
            break;
        case iLAMBDA:
          /*foutput_str("( ");
            foutput_lambda (node_lnk_1(e));
            foutput_str(" with ");
            foutput_expn(node_lnk_2(e),-1);
            foutput_str(" )");*/
            foutput_ilambda(e);
            break;    
        default:
            error_break1 ("foutput_val: expression of type %s cannot be output",
                    print_name (e));
    }
}

/*

show_state  
  
id = node_env_idlist(node_bnd_env(bnd_1)

while (id != NULL) {
   NODEPTR loc = curr_loc (mem_loc (id));

   output_str ("**Current");
   output_str (";");
   output_str (node_nm_name(id));
   output_str ("=");
   if (!mem_defined1 (loc)) {
     output_str ("?");
    }
    else {
     if (curr_val_defined (get_mem (loc))) {
       output_val_curr(get_mem (loc));
      }
      else {
       output_str ("?");
      }
    }
    newline;
    loc = next_loc (mem_loc (id));
    output_str ("**Next");
    output_str (";");
    output_str (node_nm_name(id));
    output_str ("=");
    if (!mem_defined1 (loc)) {
      output_str ("?");
    }
    else {
       if (next_val_defined (get_mem (loc))) {
         output_val_next(get_mem (loc));
       }
       else {
         output_str ("?");
       }
    }
    newline;
    for (i=2; i<=1+node_mem_history(loc); i=i+1) {
       loc = prev_loc1(mem_loc (id),i);
       printf ("%s%d","**Previous",node_mem_history(loc)+2-i);
       output_str (";");
       output_str (node_nm_name(id));
       output_str ("=");    
       if (!mem_defined1 (loc)) {
         output_str ("?");
       }
       else {
         if (prev_val_defined (get_mem (loc),i)) {
           output_val_prev(get_mem (loc),i);
         }
         else {
           output_str ("?");
         }
        }
        newline;
     }


 ***********
 #define mem_loc(X) (EnvStack[EnvBase+node_nm_level(X)+node_nm_call_level_offset(X)].env_base+node_nm_off(X))


 ************
 reduce_ref_and_fetch1: 
 if (mem_defined1 (loc)) {
   if (is_curr(loc)) {
     val = get_mem(loc);
     if (curr_val_defined(val)) {
       e = copy_curr_val(val);
     } else {
       e = val;
     }
    }
    if (is_next(loc)) {
     val = get_mem(loc);
     if (next_val_defined(val)) {
       e = copy_next_val(val);
      } else {
       e = val;
     }
    }
    for(k=2;k<=1+node_mem_history(loc);k = k+1) {
      if (is_prev1(loc,k)) {
        val= get_mem(loc);
        if (prev_val_defined(val,k)) {
          e = copy_prev_val(val,k);
        }
        else {
          e = val;
        }
      }
     }
   }
   *******************
   show_mem1
   dereference(loc);
   if (node_next(loc) != NULL){
        if (node_type(node_next(loc))==LIST) {
            if (is_curr(loc))
                output_list_curr(node_next(loc));
            if (is_next(loc))
                output_list_next(node_next(loc));
            for (i=0; i<=1+node_mem_history(loc); i=i+1)
                if (is_prev1(loc,i))
                    output_list_prev(node_next(loc),i);
        }
        else {
            output_expn (node_next(loc), -1);
        }
    }
    else {
        output_expn (node_next(loc), -1);
    }
    *****************
    show_mem_all1(NODEPTR loc)
{
    unsigned int i;

    show_mem1(curr_loc(loc));
    newline;
    show_mem1(next_loc(loc));
    newline;
    for (i=0; i<=1+node_mem_history(loc); i = i + 1) {
        show_mem1(prev_loc1(loc,i));
        newline;
    }
    
        
}

 */

void output_ilambda_prev(NODEPTR ilambda, unsigned int k)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k1;
  
  output_str("( ");
  output_lambda (node_lnk_1(ilambda));
  output_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    output_str (node_nm_name(id));
    output_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      output_expn (node_next(curr_loc(loc)), -1);
    } else {
      output_str("( Curr: ");
      output_expn (node_next(curr_loc(loc)), -1);
      output_str(", Next: ");
      output_expn (node_next(next_loc(loc)), -1);
      output_str(", Prev: ");
      for(k1=2;k<=1+node_mem_history(loc);k1 = k1+1) {
        output_str("(");
        output_int(k1-1,0);
        output_str(") ");
        output_expn (node_next(prev_loc1(loc,k1)), -1);
        if (k1 < 1+node_mem_history(loc)) {
          output_str("; ");
        }         
      }
      output_str(") ");
    }
    if (node_next(id) != NULL)
      output_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    output_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      output_str (node_nm_name(id));
      output_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        output_expn (node_next(curr_loc(loc)), -1);
      } else {
        output_str("( Curr: ");
        output_expn (node_next(curr_loc(loc)), -1);
        output_str(", Next: ");
        output_expn (node_next(next_loc(loc)), -1);
        output_str(", Prev: ");
        for(k1=2;k<=1+node_mem_history(loc);k1 = k1+1) {
          output_str("(");
          output_int(k1-1,0);
          output_str(") ");
          output_expn (node_next(prev_loc1(loc,k1)), -1);
          if (k1 < 1+node_mem_history(loc)) {
            output_str("; ");
          }         
        }
        output_str(") ");
      }
      if (node_next(id) != NULL)
        output_str (", ");
    }
  }
  output_str(" )");
}

void foutput_ilambda_prev(NODEPTR ilambda, unsigned int k)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k1;
  
  foutput_str("( ");
  foutput_lambda (node_lnk_1(ilambda));
  foutput_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    foutput_str (node_nm_name(id));
    foutput_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      foutput_expn (node_next(curr_loc(loc)), -1);
    } else {
      foutput_str("( Curr: ");
      foutput_expn (node_next(curr_loc(loc)), -1);
      foutput_str(", Next: ");
      foutput_expn (node_next(next_loc(loc)), -1);
      foutput_str(", Prev: ");
      for(k1=2;k1<=1+node_mem_history(loc);k1 = k1+1) {
        foutput_str("(");
        foutput_int(k1-1,0);
        foutput_str(") ");
        foutput_expn (node_next(prev_loc1(loc,k1)), -1);
        if (k1 < 1+node_mem_history(loc)) {
          foutput_str("; ");
        }         
      }
      foutput_str(") ");
    }
    if (node_next(id) != NULL)
      foutput_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    foutput_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      foutput_str (node_nm_name(id));
      foutput_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        foutput_expn (node_next(curr_loc(loc)), -1);
      } else {
        foutput_str("( Curr: ");
        foutput_expn (node_next(curr_loc(loc)), -1);
        foutput_str(", Next: ");
        foutput_expn (node_next(next_loc(loc)), -1);
        foutput_str(", Prev: ");
        for(k1=2;k1<=1+node_mem_history(loc);k1 = k1+1) {
          foutput_str("(");
          foutput_int(k1-1,0);
          foutput_str(") ");
          foutput_expn (node_next(prev_loc1(loc,k1)), -1);
          if (k1 < 1+node_mem_history(loc)) {
            foutput_str("; ");
          }         
        }
        foutput_str(") ");
      }
      if (node_next(id) != NULL)
        foutput_str (", ");
    }
  }
  foutput_str(" )");
}

void output_ilambda_curr(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k;
  
  output_str("( ");
  output_lambda (node_lnk_1(ilambda));
  output_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    output_str (node_nm_name(id));
    output_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (1) {
      dereference(loc);
    }
    if (node_mem_isstatic(loc)) {
      output_expn (node_next(curr_loc(loc)), -1);
    } else {
      output_str("( Curr: ");
      output_expn (node_next(curr_loc(loc)), -1);
      output_str(", Next: ");
      output_expn (node_next(next_loc(loc)), -1);
      output_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        output_str("(");
        output_int(k-1,0);
        output_str(") ");
        output_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          output_str("; ");
        }         
      }
      output_str(") ");
    }
    if (node_next(id) != NULL)
      output_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    output_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      output_str (node_nm_name(id));
      output_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        output_expn (node_next(curr_loc(loc)), -1);
      } else {
        output_str("( Curr: ");
        output_expn (node_next(curr_loc(loc)), -1);
        output_str(", Next: ");
        output_expn (node_next(next_loc(loc)), -1);
        output_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          output_str("(");
          output_int(k-1,0);
          output_str(") ");
          output_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            output_str("; ");
          }         
        }
        output_str(") ");
      }
      if (node_next(id) != NULL)
        output_str (", ");
    }
  }
  output_str(" )");
}

void foutput_ilambda_curr(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k;
  
  foutput_str("( ");
  foutput_lambda (node_lnk_1(ilambda));
  foutput_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    foutput_str (node_nm_name(id));
    foutput_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      foutput_expn (node_next(curr_loc(loc)), -1);
    } else {
      foutput_str("( Curr: ");
      foutput_expn (node_next(curr_loc(loc)), -1);
      foutput_str(", Next: ");
      foutput_expn (node_next(next_loc(loc)), -1);
      foutput_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        foutput_str("(");
        foutput_int(k-1,0);
        foutput_str(") ");
        foutput_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          foutput_str("; ");
        }         
      }
      foutput_str(") ");
    }
    if (node_next(id) != NULL)
      foutput_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    output_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      foutput_str (node_nm_name(id));
      foutput_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        foutput_expn (node_next(curr_loc(loc)), -1);
      } else {
        foutput_str("( Curr: ");
        foutput_expn (node_next(curr_loc(loc)), -1);
        foutput_str(", Next: ");
        foutput_expn (node_next(next_loc(loc)), -1);
        foutput_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          foutput_str("(");
          foutput_int(k-1,0);
          foutput_str(") ");
          foutput_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            foutput_str("; ");
          }         
        }
        foutput_str(") ");
      }
      if (node_next(id) != NULL)
        foutput_str (", ");
    }
  }
  foutput_str(" )");
}


void output_ilambda_next(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k;

  output_str("( ");
  output_lambda (node_lnk_1(ilambda));
  output_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    output_str (node_nm_name(id));
    output_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      output_expn (node_next(curr_loc(loc)), -1);
    } else {
      output_str("( Curr: ");
      output_expn (node_next(curr_loc(loc)), -1);
      output_str(", Next: ");
      output_expn (node_next(next_loc(loc)), -1);
      output_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        output_str("(");
        output_int(k-1,0);
        output_str(") ");
        output_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          output_str("; ");
        }         
      }
      output_str(") ");
    }
    if (node_next(id) != NULL)
      output_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    output_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      output_str (node_nm_name(id));
      output_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        output_expn (node_next(curr_loc(loc)), -1);
      } else {
        output_str("( Curr: ");
        output_expn (node_next(curr_loc(loc)), -1);
        output_str(", Next: ");
        output_expn (node_next(next_loc(loc)), -1);
        output_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          output_str("(");
          output_int(k-1,0);
          output_str(") ");
          output_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            output_str("; ");
          }         
        }
        output_str(") ");
      }
      if (node_next(id) != NULL)
        output_str (", ");
    }
  }
  output_str(" )");
}

void foutput_ilambda_next(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc ;
  int k;

  foutput_str("( ");
  foutput_lambda (node_lnk_1(ilambda));
  foutput_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    foutput_str (node_nm_name(id));
    foutput_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      foutput_expn (node_next(curr_loc(loc)), -1);
    } else {
      foutput_str("( Curr: ");
      foutput_expn (node_next(curr_loc(loc)), -1);
      foutput_str(", Next: ");
      foutput_expn (node_next(next_loc(loc)), -1);
      foutput_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        foutput_str("(");
        foutput_int(k-1,0);
        foutput_str(") ");
        foutput_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          foutput_str("; ");
        }         
      }
      foutput_str(") ");
    }
    if (node_next(id) != NULL)
      foutput_str (", ");
  }
  
  while (node_bnd_next(bnd_1) != NULL) {
    foutput_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      foutput_str (node_nm_name(id));
      foutput_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        foutput_expn (node_next(curr_loc(loc)), -1);
      } else {
        foutput_str("( Curr: ");
        foutput_expn (node_next(curr_loc(loc)), -1);
        foutput_str(", Next: ");
        foutput_expn (node_next(next_loc(loc)), -1);
        foutput_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          foutput_str("(");
          foutput_int(k-1,0);
          foutput_str(") ");
          foutput_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            foutput_str("; ");
          }         
        }
        foutput_str(") ");
      }
      if (node_next(id) != NULL)
        foutput_str (", ");
    }
  }
  foutput_str(" )");
}

void output_ilambda(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc;
  int k;
  
  output_str("( ");
  output_lambda (node_lnk_1(ilambda));
  output_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    output_str (node_nm_name(id));
    output_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
        output_expn (node_next(curr_loc(loc)), -1);
      } else {
      output_str("( Curr: ");
      output_expn (node_next(curr_loc(loc)), -1);
      output_str(", Next: ");
      output_expn (node_next(next_loc(loc)), -1);
      output_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        output_str("(");
        output_int(k-1,0);
        output_str(") ");
        output_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          output_str("; ");
        }         
      }
      output_str(") ");
    }
    if (node_next(id) != NULL)
      output_str (", ");
  }
  while (node_bnd_next(bnd_1) != NULL) {
    output_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      output_str (node_nm_name(id));
      output_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        output_expn (node_next(curr_loc(loc)), -1);
      } else {
        output_str("( Curr: ");
        output_expn (node_next(curr_loc(loc)), -1);
        output_str(", Next: ");
        output_expn (node_next(next_loc(loc)), -1);
        output_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          output_str("(");
          output_int(k-1,0);
          output_str(") ");
          output_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            output_str("; ");
          }         
        }
        output_str(") ");
      }
      if (node_next(id) != NULL)
        output_str (", ");
    }
  }
  output_str(" )");
}

void foutput_ilambda(NODEPTR ilambda)
{
  NODEPTR bnd_1 = node_lnk_2(ilambda);
  NODEPTR bnd_blk_base_loc;
  NODEPTR id, loc, val, loc1;
  int k;
  
  foutput_str("( ");
  foutput_lambda (node_lnk_1(ilambda));
  foutput_str(" with ");
  bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
  for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
                   id = node_next(id)) {
    foutput_str (node_nm_name(id));
    foutput_str (" = ");
    loc = bnd_blk_base_loc + node_nm_off(id);
    if (is_indirect_mem(loc)) {
      dereference(loc);
    }
    if (1) {
      foutput_expn (node_next(curr_loc(loc)), -1);
    } else {
      foutput_str("( Curr: ");
      foutput_expn (node_next(curr_loc(loc)), -1);
      foutput_str(", Next: ");
      foutput_expn (node_next(next_loc(loc)), -1);
      foutput_str(", Prev: ");
      for(k=2;k<=1+node_mem_history(loc);k = k+1) {
        foutput_str("(");
        foutput_int(k-1,0);
        foutput_str(") ");
        foutput_expn (node_next(prev_loc1(loc,k)), -1);
        if (k < 1+node_mem_history(loc)) {
          foutput_str("; ");
        }         
      }
      foutput_str(") ");
    }
    if (node_next(id) != NULL)
      foutput_str (", ");
  }
  
  while (node_bnd_next(bnd_1) != NULL) {
    foutput_str(" ; ");
    bnd_1 = node_bnd_next(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    for (id = node_env_idlist(node_bnd_env(bnd_1)); (id != NULL);
         id = node_next(id)) {
      foutput_str (node_nm_name(id));
      foutput_str (" = ");
      loc = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc)) {
        dereference(loc);
      }
      if (1) {
        foutput_expn (node_next(curr_loc(loc)), -1);
      } else {
        foutput_str("( Curr: ");
        foutput_expn (node_next(curr_loc(loc)), -1);
        foutput_str(", Next: ");
        foutput_expn (node_next(next_loc(loc)), -1);
        foutput_str(", Prev: ");
        for(k=2;k<=1+node_mem_history(loc);k = k+1) {
          foutput_str("(");
          foutput_int(k-1,0);
          foutput_str(") ");
          foutput_expn (node_next(prev_loc1(loc,k)), -1);
          if (k < 1+node_mem_history(loc)) {
            foutput_str("; ");
          }         
        }
        foutput_str(") ");
      }
      if (node_next(id) != NULL)
        foutput_str (", ");
    }
  }
  foutput_str(" )");
}






void output_list_prev(NODEPTR list, unsigned int k)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = prev_loc1(list_base (list),k);
    size = mem_defined1 (loc = prev_loc1(list_size (list),k)) ? get_int (loc) : prev_max_size (list,k);
        /*newline;*/
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            output_val_prev (get_mem (prev_loc1(loc,k)),k);
        }
        else {
            output_char ('?');
        }
        if ((int)++i < size)
            output_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    output_char (']');
}

void foutput_list_prev(NODEPTR list, unsigned int k)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = prev_loc1(list_base (list),k);
    size = mem_defined1 (loc = prev_loc1(list_size (list),k)) ? get_int (loc) : prev_max_size (list,k);
        /*newline;*/
    foutput_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            foutput_val_prev (get_mem (prev_loc1(loc,k)),k);
        }
        else {
            foutput_char ('?');
        }
        if ((int)++i < size)
            foutput_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    foutput_char (']');
}


void output_list_next(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = next_loc(list_base (list));
    size = mem_defined1 (loc = next_loc(list_size (list))) ? get_int (loc) : next_max_size (list);
        /*newline;*/
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            output_val_next (get_mem (next_loc(loc)));
        }
        else {
            output_char ('?');
        }
        if ((int)++i < size)
            output_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    output_char (']');
}

void foutput_list_next(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = next_loc(list_base (list));
    size = mem_defined1 (loc = next_loc(list_size (list))) ? get_int (loc) : next_max_size (list);
        /*newline;*/
    foutput_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            foutput_val_next (get_mem (next_loc(loc)));
        }
        else {
            foutput_char ('?');
        }
        if ((int)++i < size)
            foutput_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    foutput_char (']');
}


void output_list_curr(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = curr_loc(list_base (list));
    size = mem_defined1 (loc = curr_loc(list_size (list))) ? get_int (loc) : curr_max_size (list);
        /*newline;*/
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            output_val_curr (get_mem (curr_loc(loc)));
        }
        else {
            output_char ('?');
        }
        if ((int)++i < size)
            output_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    output_char (']');
}

void foutput_list_curr(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = curr_loc(list_base (list));
    size = mem_defined1 (loc = curr_loc(list_size (list))) ? get_int (loc) : curr_max_size (list);
        /*newline;*/
    foutput_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            foutput_val_curr (get_mem (curr_loc(loc)));
        }
        else {
            foutput_char ('?');
        }
        if ((int)++i < size)
            foutput_char (',');
            /*newline;
              show_mem_all1(loc);*/
    }
    foutput_char (']');
}

void output_list(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;
    base_loc = list_base (list);
    size = mem_defined1 (loc = list_size (list)) ? get_int (loc) : max_size (list);
    output_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc)))) {
            output_val (get_mem (loc));
        }
        else {
            output_char ('?');
        }
        if ((int)++i < size)
            output_char (',');
    }
    output_char (']');
}


void foutput_list(NODEPTR list)
{
    int     size;
    unsigned i = 0;
    NODEPTR loc,
            base_loc;

    base_loc = list_base (list);
    size = mem_defined1 (loc = list_size (list)) ? get_int (loc) : max_size (list);
    foutput_char ('[');
    while ((int)i < size) {
        if (mem_defined1 (loc = base_loc + offset2 (i, node_mem_history(loc))))
            foutput_val (get_mem (loc));
        else
            foutput_char ('?');
        if ((int)++i < size)
            foutput_char (',');
    }
    foutput_char (']');
}


/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_val(NODEPTR val)
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:  
        case MPFR:  
        case LOC:
        case LAMBDA:  
        case STRUCT:
            return (val);
        case iLAMBDA:
          return(copy_ilambda(val));
          /*return(node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val))); */
          /*return (val);*/
        case LIST:
            return (copy_list(val));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}

/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_prev_val(NODEPTR val, unsigned int k)
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:  
        case MPFR:  
        case LOC:
        case LAMBDA:  
        case STRUCT:
            return (val);
        case iLAMBDA:
          return(copy_prev_ilambda(val,k));  
          /*return(node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val))); */
          /*return (val);*/
        case LIST:
            return (prev_copy_list(val,k));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}

/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_curr_val(NODEPTR val)
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:
        case LOC:
        case LAMBDA:  
        case STRUCT:
            return (val);
        case iLAMBDA:
          return(copy_curr_ilambda(val));  
          /*return(node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val))); */
          /*return (val);*/
        case LIST:
            return (curr_copy_list(val));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}

/* Return a copy of a value. Atoms need no copying. */
NODEPTR copy_next_val(NODEPTR val)
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LOC:
        case LAMBDA:  
        case STRUCT:
            return (val);
        case iLAMBDA:
          return(copy_next_ilambda(val)); 
          /*return(node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val))); */
          /*return (val);*/
        case LIST:
            return (next_copy_list(val));
        default:
            error_break1 ("cannot copy value of type %s", print_name(val));
            return (F);
    }
}




void copy_indirect_mem(NODEPTR loc, register NODEPTR val)
{
    unsigned int i;
    NODEPTR cloc, nloc, ploc, pnloc ;
    dereference (val);
    
    cloc = curr_loc (loc);
    nloc = next_loc (loc);
    ploc = prev_loc1(loc,0);
    pnloc = prev_loc1(loc,1);
    /*
    printf("copy_indirect: is_curr = %d, is_next = %d\n",is_curr(val),is_next(val));
    */
    /*if (is_curr(val)) {*/
      node_next(cloc) = curr_loc(val);
      node_next(nloc) = next_loc (val);
      node_next(ploc) = prev_loc1(val,0);
      node_next(pnloc) = prev_loc1 (val,1);
      node_mem_isindirect(cloc) = node_mem_isindirect(nloc) = T;
      node_mem_isindirect(ploc) = node_mem_isindirect(pnloc) = T;
      for (i=2;i<=1+node_mem_history(loc);i = i+1) {
        node_next(prev_loc1(loc,i)) = prev_loc1(val,i);
        node_mem_isindirect(prev_loc1(loc,i)) = T;
      }
      /*
    }
      */
      /*  
    if (is_next(val)) {
      node_next(cloc) = curr_loc(val);
      node_next(nloc) = next_loc (val);
      node_next(ploc) = prev_loc1(val,0);
      node_next(pnloc) = prev_loc1 (val,1);
      node_mem_isindirect(cloc) = node_mem_isindirect(nloc) = T;
      node_mem_isindirect(ploc) = node_mem_isindirect(pnloc) = T;
      for (i=2;i<=1+node_mem_history(loc);i = i+1) {
        node_next(prev_loc1(loc,i)) = prev_loc1(val,i);
        node_mem_isindirect(prev_loc1(loc,i)) = T;
      }
    }
      */
    
}


/* make a copy of an ilambda */
/* the copy is defined in the CURRENT state */
NODEPTR copy_ilambda(NODEPTR ilambda)
{
  unsigned int history = 1;
  NODEPTR base_loc, bnd_blk_base_loc;
  NODEPTR bnd_1, bnd_new, bnd_new_1;
  NODEPTR env, bnd_env;
  NODEPTR blk;
  NODEPTR id;
  NODEPTR lambda_new;
  NODEPTR loc, loc_old;
  
  bnd_1 = node_lnk_2(ilambda);

  if (bnd_1 != NULL) {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    /*history = node_env_history(bnd_env);*/
    /*printf("copying ilambda\n");*/
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (node_mem_history(loc_old) > history) {
                history = node_mem_history(loc_old);
            }
    }
    blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
    base_loc = node_blk_base(blk);
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc_old)) {
        loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
        copy_indirect_mem(loc, loc_old);
      } else {
        loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
        if (mem_defined1(loc_old)) {
          put_mem1 (loc,get_mem(loc_old));
        }
      }
    }

    bnd_new = bnd_node(bnd_env, blk, NULL); 
    bnd_1 = node_bnd_next(bnd_1);
    bnd_new_1 = bnd_new;

    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
      /*history = node_env_history(bnd_env); */
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id);
        if (node_mem_history(loc_old) > history) {
                  history = node_mem_history(loc_old);
        }
      }
      blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
      base_loc = node_blk_base(blk);
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id); 
        if (is_indirect_mem(loc_old)) {
          loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
          copy_indirect_mem(loc, loc_old);
        } else {
          loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
          if (mem_defined1(loc_old)) {
            put_mem1 (loc,get_mem(loc_old));
          }
        }
      }

      node_bnd_next(bnd_new_1) = bnd_node(bnd_env, blk, NULL);
      bnd_1 = node_bnd_next(bnd_1);
      bnd_new_1 = node_bnd_next(bnd_new_1);
    }

    lambda_new = node2(iLAMBDA, node_lnk_1(ilambda),bnd_new);
    /*node_lnk_3(lambda) = bnd_new;*/

  } else {
    lambda_new = ilambda;
  }
  
  return (lambda_new);
}




/* make a copy of a lambda */
/* the copy is defined in the CURRENT state */
NODEPTR copy_curr_ilambda(NODEPTR ilambda)
{
  unsigned int history = 1;
  NODEPTR base_loc, bnd_blk_base_loc;
  NODEPTR bnd_1, bnd_new, bnd_new_1;
  NODEPTR env, bnd_env;
  NODEPTR blk;
  NODEPTR id;
  NODEPTR lambda_new;
  NODEPTR loc, loc_old;
  
  bnd_1 = node_lnk_2(ilambda);

  if (bnd_1 != NULL) {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = (node_blk_base(node_bnd_blk(bnd_1)));
    /*history = node_env_history(bnd_env);*/
    /*printf("copying curr_ilambda \n");*/
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (node_mem_history(loc_old) > history) {
                history = node_mem_history(loc_old);
            }
    }
    blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
    base_loc = node_blk_base(blk);
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc_old)) {
        loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
        /*
        printf("old:\n");
        show_mem_all1(loc_old);
        printf("new:\n");
        */
        copy_indirect_mem(loc, loc_old);
        /*
        show_mem_all1(loc);
        */
      } else {
        loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
         if (mem_defined1(loc_old)) {
           put_mem1 (loc, get_mem(loc_old));
         }
      }
    }

    bnd_new = bnd_node(bnd_env, blk, NULL); 
    bnd_1 = node_bnd_next(bnd_1);
    bnd_new_1 = bnd_new;

    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
      /*history = node_env_history(bnd_env); */
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id);
        if (node_mem_history(loc_old) > history) {
                  history = node_mem_history(loc_old);
        }
      }
      blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
      base_loc = node_blk_base(blk);
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id); 
        if (is_indirect_mem(loc_old)) {
          loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
          copy_indirect_mem(loc, loc_old);
        } else {
          loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
          if (mem_defined1(loc_old)) {
            put_mem1 (loc, get_mem(loc_old));
          }
        }
      }

      node_bnd_next(bnd_new_1) = bnd_node(bnd_env, blk, NULL);
      bnd_1 = node_bnd_next(bnd_1);
      bnd_new_1 = node_bnd_next(bnd_new_1);
    }

    lambda_new = node2(iLAMBDA, node_lnk_1(ilambda), bnd_new);
      /*node_lnk_3(lambda) = bnd_new;*/
    
  } else {
    lambda_new = ilambda;
  }
  
  return (lambda_new);
}

/* make a copy of a lambda */
/* the copy is defined in the CURRENT state */
NODEPTR copy_next_ilambda(NODEPTR ilambda)
{
  unsigned int history = 1;
  NODEPTR base_loc, bnd_blk_base_loc;
  NODEPTR bnd_1, bnd_new, bnd_new_1;
  NODEPTR env, bnd_env;
  NODEPTR blk;
  NODEPTR id;
  NODEPTR lambda_new;
  NODEPTR loc, loc_old;
 

  
  bnd_1 = node_lnk_2(ilambda);

  if (bnd_1 != NULL) {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    /*history = node_env_history(bnd_env);*/
    /*printf("copying next_ilambda\n");*/
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (node_mem_history(loc_old) > history) {
                history = node_mem_history(loc_old);
            }
    }
    blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
    base_loc = node_blk_base(blk);
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc_old)) {
        loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
        copy_indirect_mem(loc, loc_old);
      } else {
        loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
        if (mem_defined1(loc_old)) {
          put_mem1 (loc,get_mem(loc_old));
        }
      }
    }

    bnd_new = bnd_node(bnd_env, blk, NULL); 
    bnd_1 = node_bnd_next(bnd_1);
    bnd_new_1 = bnd_new;

    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
      /*history = node_env_history(bnd_env); */
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id);
        if (node_mem_history(loc_old) > history) {
                  history = node_mem_history(loc_old);
        }
      }
      blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
      base_loc = node_blk_base(blk);
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id); 
        if (is_indirect_mem(loc_old)) {
          loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
          copy_indirect_mem(loc, loc_old);
        } else {
          loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
          if (mem_defined1(loc_old)) {
            put_mem1 (loc,get_mem(loc_old));
          }
        }
      }

      node_bnd_next(bnd_new_1) = bnd_node(bnd_env, blk, NULL);
      bnd_1 = node_bnd_next(bnd_1);
      bnd_new_1 = node_bnd_next(bnd_new_1);
    }

    lambda_new = node2(iLAMBDA, node_lnk_1(ilambda), bnd_new);
    /*node_lnk_3(lambda) = bnd_new;*/
  } else {
    lambda_new = ilambda;
  }
  
  return (lambda_new);
}

/* make a copy of a lambda */
/* the copy is defined in the CURRENT state */
NODEPTR copy_prev_ilambda(NODEPTR ilambda, unsigned int k)
{
  unsigned int history = 1;
  NODEPTR base_loc, bnd_blk_base_loc;
  NODEPTR bnd_1, bnd_new, bnd_new_1;
  NODEPTR env, bnd_env;
  NODEPTR blk;
  NODEPTR id;
  NODEPTR lambda_new;
  NODEPTR loc, loc_old;

  
  bnd_1 = node_lnk_2(ilambda);

  if (bnd_1 != NULL) {
    bnd_env = node_bnd_env(bnd_1);
    bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
    /*history = node_env_history(bnd_env);*/
    /*printf("copying prev_ilambda\n");*/
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (node_mem_history(loc_old) > history) {
                history = node_mem_history(loc_old);
            }
    }
    blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
    base_loc = node_blk_base(blk);
    for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
      loc_old = bnd_blk_base_loc + node_nm_off(id);
      if (is_indirect_mem(loc_old)) {
        loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
        copy_indirect_mem(loc, loc_old);
      } else {
        loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
        if (mem_defined1(loc_old)) {
          put_mem1 (loc, get_mem(loc_old));
        }
      }
    }

    bnd_new = bnd_node(bnd_env, blk, NULL); 
    bnd_1 = node_bnd_next(bnd_1);
    bnd_new_1 = bnd_new;

    while (bnd_1 != NULL) {
      bnd_env = node_bnd_env(bnd_1);
      bnd_blk_base_loc = node_blk_base(node_bnd_blk(bnd_1));
      /*history = node_env_history(bnd_env); */
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id);
        if (node_mem_history(loc_old) > history) {
                  history = node_mem_history(loc_old);
        }
      }
      blk = mem_block2 (STATIC, node_env_size(bnd_env), history, NoFraming);
      base_loc = node_blk_base(blk);
      for (id = node_env_idlist(bnd_env); (id != NULL); id = node_next(id)) {
        loc_old = bnd_blk_base_loc + node_nm_off(id); 
        if (is_indirect_mem(loc_old)) {
          loc = new_loc2 (base_loc + node_nm_off(id), node_mem_isstatic(loc_old), node_mem_history(loc_old), is_framed_mem(loc_old));
          copy_indirect_mem(loc, loc_old);
        } else {
          loc = new_loc2 (base_loc + node_nm_off(id), STATIC, 1, 0);
           if (mem_defined1(loc_old)) {
            put_mem1 (loc,get_mem(loc_old));
           }
        }
      }

      node_bnd_next(bnd_new_1) = bnd_node(bnd_env, blk, NULL);
      bnd_1 = node_bnd_next(bnd_1);
      bnd_new_1 = node_bnd_next(bnd_new_1);
    }

    lambda_new = node2(iLAMBDA, node_lnk_1(ilambda), bnd_new);
    /*node_lnk_3(lambda) = bnd_new;*/

  } else {
    lambda_new = ilambda;
  }
  
  return (lambda_new);
}


/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR copy_list(NODEPTR list)
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    unsigned int history = node_mem_history(size_loc);
    unsigned framed = is_framed_mem(size_loc);
    
    assert0 (list_defined (list),
            "attempt to copy an incompletely defined list");
    
    list1 = new_list2 (STATIC, STATIC, get_mem (size_loc), history, framed);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (size_loc);
    for (i = 0; i < size; i += 1)
      if (mem_defined1(base_loc + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i,history),
                  copy_val (get_mem (base_loc + offset2 (i,history))));
    return (list);
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR prev_copy_list(NODEPTR list, unsigned int k)
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    unsigned int history = node_mem_history(size_loc);
    unsigned framed = is_framed_mem(size_loc);
    NODEPTR psize_loc = prev_loc1(size_loc,k);
    NODEPTR pbase_loc = prev_loc1(base_loc,k);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    
    assert0 (prev_list_defined (list,k),
            "attempt to copy an incompletely defined list");
    
    list1 = new_list2 (STATIC, STATIC, get_mem (psize_loc), history, framed);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (psize_loc);
    for (i = 0; i < size; i += 1)
      if (mem_defined1(pbase_loc + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i,history),
                  copy_prev_val (get_mem (pbase_loc + offset2 (i,history)),k));
    return (list);
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR curr_copy_list(NODEPTR list)
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    unsigned int history = node_mem_history(size_loc);
    unsigned framed = is_framed_mem(size_loc);
    NODEPTR csize_loc = curr_loc(size_loc);
    NODEPTR cbase_loc = curr_loc(base_loc);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    
    assert0 (curr_list_defined (list),
            "attempt to copy an incompletely defined list");
    
    list1 = new_list2 (STATIC, STATIC, get_mem (csize_loc), history, framed);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (csize_loc);
    for (i = 0; i < size; i += 1)
      if (mem_defined1(cbase_loc + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i,history),
                  copy_curr_val (get_mem (cbase_loc + offset2 (i,history))));
    return (list);
}

/* Make a copy of a list. */
/* The copy is defined in the CURRENT state. */
NODEPTR next_copy_list(NODEPTR list)
{
    NODEPTR size_loc = list_size (list);
    NODEPTR base_loc = list_base (list);
    unsigned int history = node_mem_history(size_loc);
    unsigned framed = is_framed_mem(size_loc);
    NODEPTR nsize_loc = next_loc(size_loc);
    NODEPTR nbase_loc = next_loc(base_loc);
    NODEPTR new_base;
    NODEPTR list1;
    int i, size;
    
    assert0 (next_list_defined (list),
            "attempt to copy an incompletely defined list");
    
    list1 = new_list2 (STATIC, STATIC, get_mem (nsize_loc), history, framed);
/*    if (is_static_mem (base_loc) && is_static_mem (size_loc)) {*/
/*printf("***SAVED\n");*/
/*        return (list);*/
/*	}*/
    list = list1;
    new_base = curr_loc (list_base (list));
    size = get_int (nsize_loc);
    for (i = 0; i < size; i += 1)
      if (mem_defined1(nbase_loc + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i,history),
                  copy_next_val (get_mem (nbase_loc + offset2 (i,history))));
    return (list);
}




int string_size(NODEPTR s)
{
    return (strlen (node_strval(s)));
}

NODEPTR substring(NODEPTR s, int i, int j)
{
    return (string_node (substr (node_strval(s), i, j)));
}

NODEPTR append_string (NODEPTR s1, NODEPTR s2)
{
    return (string_node (strapnd (node_strval(s1), node_strval(s2))));
}

/* Return the result of appending list2 to the end of list1. */
/* The new list is defined in the CURRENT state. */
NODEPTR append_list(NODEPTR list1, NODEPTR list2)
{
    NODEPTR     new_base,
            base_loc1,
            base_loc2;
    int     size1,
            size2;
    unsigned i;
    unsigned int history;
    unsigned framed;
        /*,history1,history2;*/
    

    base_loc1 = list_base (list1);
    history = node_mem_history(list_size (list1));
    framed = is_framed_mem(list_size (list1));
    size1 = get_int (list_size (list1));
    base_loc2 = list_base (list2);
        /*history2 = node_mem_history(list_size (list2));*/
    size2 = get_int (list_size (list2));
        /*if (history1<history2)
        history=history1;
    else
        history=history2;
    if (history<1)  history=1;
    if (history>MAXHISTORY) history=MAXHISTORY;*/
    list1 = new_list2 (STATIC, STATIC, int_node ((unsigned int)size1 + size2), history, framed);
    new_base = curr_loc (node_lst_base(list1));
    for (i = 0; (int)i < size1; i += 1) {
      if (mem_defined1(base_loc1 + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i, history),
                  copy_val (get_mem (base_loc1 + offset2 (i,history))));
    }
    new_base += offset2 (i,history);
    for (i = 0; (int)i < size2; i += 1) {
      if (mem_defined1(base_loc2 + offset2 (i,history)))
        put_mem1 (new_base + offset2 (i,history),
                  copy_val (get_mem (base_loc2 + offset2 (i,history))));
    }
    return (list1);
}

/* Create a new list. */
/*
NODEPTR new_list(cell_type,size_type,n)
int cell_type,size_type;
NODEPTR n;
{
    NODEPTR blk = mem_block (cell_type, (unsigned)node_ival(n) + 1);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = size_loc + offset (1);

    new_loc (size_loc, size_type);
    if (size_type == STATIC)
        put_mem (size_loc, n);
    return (list_node (blk, size_loc, base_loc));
}
*/
/* Create a new list. */
/*
NODEPTR new_list1(cell_type,size_type,n,history)
int cell_type,size_type;
NODEPTR n;
unsigned int history;
{
    NODEPTR blk = mem_block1 (cell_type, (unsigned int)node_ival(n) + 1,history);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = (node_ival(n)==0)? size_loc : size_loc + offset2 (1,history);
        
    new_loc1 (size_loc, size_type, history);
    if (size_type == STATIC)
        put_mem1 (size_loc, n);
    return (list_node (blk, size_loc, base_loc));
}
*/

/* Create a new list. */
NODEPTR new_list2(int cell_type, int size_type, NODEPTR n, unsigned int history, unsigned framed)
{
    NODEPTR blk = mem_block2 (cell_type, (unsigned int)node_ival(n) + 1,history,framed);
    NODEPTR size_loc = node_blk_base(blk),
        base_loc = (node_ival(n)==0)? size_loc : size_loc + offset2 (1,history);
        
    new_loc2 (size_loc, size_type, history, framed);
    if (size_type == STATIC)
        put_mem1 (size_loc, n);
    return (list_node (blk, size_loc, base_loc));
}

/* Assign a value to a location */
/*
void assign_val (loc,val)
NODEPTR val;
register NODEPTR loc;
{
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LAMBDA:
        case LOC:
            if (mem_defined (loc)) {
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else{
                        if (node_mem_isundefined(loc)){
                            printf("do we want this?\n");
                                put_mem (loc, val);
                        }
                        else{
                            error_break0 ("attempt to overwrite variable");
                        }
                    }
		} 
            }
            else
                put_mem (loc, val);
            break;
        case STRUCT:
            assign_struct (loc, node_lnk_1(val));
            break;
        case LIST:
            assign_list (loc, val);
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}
*/

/* Assign a value to a location */
void assign_val1 (NODEPTR loc, register NODEPTR val)
{
        /*if (DebugLevel>2)
          printf("assign_val: history=%d\n",node_mem_history(loc));*/
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA:
        case iLAMBDA:  
        case LOC:
            if (mem_defined1 (loc)) {
                
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else{
                        if (node_mem_isundefined(loc)){
                            printf("do we want this?\n");
                                /*put_mem1 (loc, val);*/
                        }
                        else{
                            assert0 (!is_list (get_list (loc)),
                         "attempt to write non-list value to a list");
                            error_break1a ("attempt to overwrite variable",get_mem (loc));
                        }
                    }
		} 
            }
            else {
                put_mem1 (loc, val);
            }
            break;
        case STRUCT:
            assign_struct (loc, node_lnk_1(val));
            break;
        case LIST:
            assign_list (loc, val);
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}

/* Assign a value to a location */
void assign_val_update (register NODEPTR loc, NODEPTR val)
{
        /*if (DebugLevel>2)
          printf("assign_val: history=%d\n",node_mem_history(loc));*/
    switch (node_type(val)) {
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case MPZ:
        case MPX:
        case MPFR:  
        case LAMBDA:
        case iLAMBDA:  
        case LOC:
            if (mem_defined1 (loc)) {
                
                if (RedundancyFlag || !equal_val (val, get_mem (loc))) {
		    if (loc == TempLoc)
		        error_break0 ("attempt to redefine constant");
		    else{
                        if (node_mem_isundefined(loc)){
                            printf("do we want this?\n");
                                /*put_mem1 (loc, val);*/
                        }
                        else{
                            assert0 (!is_list (get_list (loc)),
                         "attempt to write non-list value to a list");
                            error_break1a ("attempt to overwrite variable",get_mem (loc));
                        }
                    }
		} 
            }
            else {
                put_mem1 (loc, val);
            }
            break;
        case STRUCT:
            assign_struct (loc, node_lnk_1(val));
            break;
        case LIST:
            assign_list_update (loc, val);
            break;
        default:
            error_break1 ("cannot assign value of type %s", print_name (val));
    }
}

void assign_struct(NODEPTR loc, NODEPTR struct1)
{   
    if (!mem_defined1 (loc)) {
        put_mem1 (loc, struct1);
    }
    else {
	register int size2,size1;
        register NODEPTR loc1;
        register NODEPTR loc2;
        NODEPTR struct2 = get_list (loc);
        unsigned int i;
        unsigned int up=3+node_mem_history(loc);
        
        type_check (struct2, LIST);
            /*printf("before putlist\n");
              show_mem_all1(loc);*/
        struct1 = put_list (loc, struct1);
            /*printf("after putlist\n");
              show_mem_all1(loc);*/

            /*printf("is_current_loc = %d\n", is_curr(loc));
              printf("ms1 = %d ms2 = %d \n",max_size (struct1),max_size (struct2));*/
        
        assert0 (max_size (struct1) >= max_size (struct2),
	         "attempt to redefine max. size of list");
        if (mem_defined1 (node_lst_size(struct2))) {
	    size2 = get_int (node_lst_size(struct2));
            if (mem_defined1 (node_lst_size(struct1))) {
                size1 = get_int (node_lst_size(struct1));
                    /*printf("s1 =%d \n",size1);*/
            }
                /*printf("s2 =%d \n",size2);*/
            if (size1 != size2) {
                for (i=0; i <= up; i = i + 1) {
                    node_mem_isstatic(node_lst_size(struct1)+i) =0;
                }
            }
            
            
	    if (node_lst_size(struct1) != node_lst_size(struct2)) {
                if (is_curr(node_lst_size(struct2))) {
                        /*printf("is_curr\n");*/
                    put_mem1 (curr_loc(node_lst_size(struct1)), get_mem (node_lst_size(struct2)));
                        /*make_indirect_mem (node_lst_size(struct2), curr_loc(node_lst_size(struct1)));*/
                }
                if (is_next(node_lst_size(struct2))) {
                        /*printf("is_next\n");*/
                    put_mem1 (next_loc(node_lst_size(struct1)), get_mem (node_lst_size(struct2)));
                        /*make_indirect_mem (node_lst_size(struct2), next_loc(node_lst_size(struct1)));*/
                        /*printf("size=%ld\n",node_ival(get_mem (node_lst_size(struct2))));*/
                    
                }
                
	    }
                /*printf("before\n");
                  show_mem_all1(loc);*/
            loc1 = node_lst_base(struct1);
            loc2 = node_lst_base(struct2);
	    if (loc1 != loc2) {
                while (size2-- > 0) {
                        /*printf("mem_defined1=%d, mem_defined2=%d\n",mem_defined1 (loc1),mem_defined1 (loc2));*/
	            if (mem_defined1 (loc2))
                        assign_val1 (loc1, get_mem (loc2));
                        /*printf("loop size2=%d\n",size2);
                          show_mem_all1(loc);*/
                        /*make_indirect_mem (loc2, loc1);*/
                        /*show_mem_all1(loc);*/
                    loc1 += offset2(1, node_mem_history(loc1));
                    loc2 += offset2(1, node_mem_history(loc2));
		}
	    }
                /*
            printf("after\n");
            show_mem_all1(loc);*/
            
	}
    }
}

/* copy one list into another. */
void assign_list(NODEPTR loc, NODEPTR list)
{
    int     current_size1,
            current_size2;
    register NODEPTR loc1;
        /*register NODEPTR ploc1;*/
    register NODEPTR loc2;
    NODEPTR size_loc1;
    NODEPTR size_loc2;
    NODEPTR list1;
    
    loc2 = list_base (list);
    size_loc2 = list_size (list);
    current_size2 = get_int (size_loc2);
    if (mem_defined1 (loc)) {
        assert0 (is_list (list1 = get_list (loc)),
                "attempt to overwrite non-list with list");
        assert2d (max_size (list1) >= current_size2,
                "attempt to assign list of size %d to one of max size %d",
                current_size2, max_size (list1));
    }
    else {
        list1 = new_list2 (is_static_mem (loc) ? STATIC : STATE,
                           STATIC, get_mem (size_loc2), node_mem_history(loc), is_framed_mem(loc));
        list1 = put_list (loc, list1);/* possibly adjusted. */
    }
    
    loc1 = list_base (list1);
        /*if (is_next(loc)) {
        ploc1 = list_base(get_list(prev_loc1(loc,1)));
    }
    else {
        ploc1 = list_base(get_list(prev_loc1(loc,0)));
        }*/
    
    size_loc1 = list_size (list1);

    if (!mem_defined1 (size_loc1)) {
        current_size1 = current_size2;
        put_mem1 (size_loc1, int_node (current_size1));
    }
    else {
        assert2d ((current_size1 = get_int (size_loc1)) == current_size2,
                "attempt to assign list of size %d to one of size %d",
                current_size2, current_size1);
    }
    while (current_size2-- > 0) {
        if (mem_defined1(loc2)) {
            assign_val1 (loc1, get_mem (loc2));
        }
        
	loc1 += offset2 (1,node_mem_history(loc1));
	loc2 += offset2 (1,node_mem_history(loc2));
    }
}

/* copy one list into another. */
void assign_list_update(NODEPTR loc, NODEPTR list)
{
    int     current_size1,
            current_size2;
    register NODEPTR loc1;
        /*register NODEPTR ploc1;*/
    register NODEPTR loc2;
    NODEPTR size_loc1;
    NODEPTR size_loc2;
    NODEPTR list1;
    
    loc2 = list_base (list);
    size_loc2 = list_size (list);
    current_size2 = get_int (size_loc2);
    if (mem_defined1 (loc)) {
        assert0 (is_list (list1 = get_list (loc)),
                "attempt to overwrite non-list with list");
        assert2d (max_size (list1) >= current_size2,
                "attempt to assign list of size %d to one of max size %d",
                current_size2, max_size (list1));
    }
    else {
        list1 = new_list2 (is_static_mem (loc) ? STATIC : STATE,
                           STATIC, get_mem (size_loc2), node_mem_history(loc), is_framed_mem(loc));
        list1 = put_list (loc, list1);/* possibly adjusted. */
    }
    
    loc1 = list_base (list1);
    size_loc1 = list_size (list1);

    if (!mem_defined1 (size_loc1)) {
        current_size1 = current_size2;
        put_mem1 (size_loc1, int_node (current_size1));
    }
    else {
        assert2d ((current_size1 = get_int (size_loc1)) == current_size2,
                "attempt to assign list of size %d to one of size %d",
                current_size2, current_size1);
    }
    while (current_size2-- > 0) {
        if (mem_defined1(loc2)) {
            if (mem_defined1(loc1)) {
                if (!val_defined(get_mem(loc1)))
                    assign_val_update (loc1, get_mem (loc2));
            }
            else
                assign_val_update (loc1, get_mem (loc2));
        }
        
	loc1 += offset2 (1,node_mem_history(loc1));
	loc2 += offset2 (1,node_mem_history(loc2));
    }
}

/* Test whether two values are equal. */
int equal_val(NODEPTR e1, NODEPTR e2)
{
    if (node_type(e1) != node_type(e2))
        return (F);
    return (compare_val (e1, e2) == 0);
}


/* Compare two values. Return 0 if e1=e2, >0 if e1>e2, <0 if e1<e2. */
int compare_val(NODEPTR e1, NODEPTR e2)
{
    unsigned type;
    int tmp;
    

    if ((type = node_type(e1)) != node_type(e2))
        return (-1);
    switch (type) {
        case NIL:
	    return (0);
        case BOOL:
        case INT:
            return (node_ival(e1) - node_ival(e2));
        case LAMBDA:
            return (e1 - e2);
        case LOC:
                  return (compare_ptrs (e1, e2));
        case STRING:
            return (compare_strings (e1, e2));
        case FLOAT:
            if (node_flval(e1) == node_flval(e2)) tmp=0;
            else if (node_flval(e1) > node_flval(e2)) tmp=1;
                    else tmp = -1;
            return (tmp);
        case MPZ:
            return (mpz_cmp(node_mpzval(e1), node_mpzval(e2)));
        case MPX:
            return (mpz_cmp(node_mpxval(e1), node_mpxval(e2)));    
        case MPFR:
            return (mpfr_cmp(node_mpfrval(e1), node_mpfrval(e2)));
        case LIST:
            return (compare_lists (e1, e2));
        default:
            error_break1 ("cannot compare values of type %s", print_name (e1));
            return (F);
    }
}

int compare_strings(NODEPTR e1, NODEPTR e2)
{
    return (strcmp (node_strval(e1), node_strval(e2)));
}

int compare_ptrs(NODEPTR e1, NODEPTR e2)
{
    return (node_loc_tonext(e1) == node_loc_tonext(e2) ?
            node_loc_ptr(e1) - node_loc_ptr(e2) :
            node_loc_ptr(e1) - node_loc_ptr(e2));
}

int compare_lists(NODEPTR list1, NODEPTR list2)
{
    int     size1,
            size2,
            answer;
    unsigned i;
    NODEPTR base_loc1,
            base_loc2;

    assert0 (list_defined (list1) && list_defined (list2),
            "attempt to compare an undefined list");
    base_loc1 = list_base (list1);
    base_loc2 = list_base (list2);
    size1 = get_int (list_size (list1));
    size2 = get_int (list_size (list2));
    if (size1 != size2)
        return (size1 - size2);
    for (i = 0; (int)i < size1; i += 1) {
        if ((answer = compare_val (get_mem (base_loc1 + offset2 (i,node_mem_history(base_loc1))),
                                            get_mem (base_loc2 + offset2 (i,node_mem_history(base_loc2))))) != 0)
            return (answer);
    }
    return (0);
}

